from typing import Any

__all__ = (
    "MaterialLocation",
    "MaterialOrLocation",
    "MaterialProcessingContext",
)


MaterialLocation = Any
MaterialOrLocation = Any
MaterialProcessingContext = Any
