from typing import Any

__all__ = (
    "BlueprintLocation",
    "BlueprintOrLocation",
    "BlueprintProcessingContext",
)


BlueprintLocation = Any
BlueprintOrLocation = Any
BlueprintProcessingContext = Any
